/*
 * Mbuni - Open  Source MMS Gateway 
 * 
 * MMSC CFG: MMC configuration and misc. functions
 * 
 * Copyright (C) 2003 - 2005, Digital Solutions Ltd. - http://www.dsmagic.com
 *
 * Paul Bagyenda <bagyenda@dsmagic.com>
 * 
 * This program is free software, distributed under the terms of
 * the GNU General Public License, with a few exceptions granted (see LICENSE)
 */
#include <sys/file.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <errno.h>
#include <dlfcn.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include "mmsc_cfg.h"
#include "mms_queue.h"

#define MMS_PORT 8191  /* Default content fetch port. */



static void *load_module(mCfgGrp *grp, char *config_key, char *symbolname)
{
     Octstr *s;
     void *retval = NULL;
     
     s = mms_cfg_get(grp, octstr_imm(config_key));
     if (s) {
	  void *x = dlopen(octstr_get_cstr(s), RTLD_LAZY);
	  void *y = NULL;
#ifdef __APPLE__
	  char sbuf[512];
	  
	  sprintf(sbuf, "_%s", symbolname);
#endif  
	  if (x == NULL || ((y = dlsym(x, symbolname)) == NULL 
#ifdef __APPLE__ /* fink version of dlsym has issues it seems. */
	      && (y = dlsym(x, sbuf)) == NULL
#endif
	       ))
	       
	       panic(0,
		     "Error, unable to load dynamic libary (%s): "
		     "libhandle is %s, funcs is %s, err=%s",
		     octstr_get_cstr(s),
		     x ? "OK" : "Not OK", y ? "OK" : "Not OK", dlerror());
	  else
	       retval = y;
	  
	  octstr_destroy(s);
     }
     return retval;
}

MmscSettings *mms_load_mmsc_settings(mCfg *cfg, List **proxyrelays)
{
     Octstr *s;

     List *l;
     mCfgGrp *grp = mms_cfg_get_single(cfg, octstr_imm("mbuni"));
     mCfgGrp *cgrp = mms_cfg_get_single(cfg, octstr_imm("core"));
     MmscSettings *m = gw_malloc(sizeof *m);
     long port = -1;
     Octstr  *from,  *user, *pass;
     Octstr *qdir = NULL;
     int i, n;

     memset(m, 0, sizeof *m);

     if (grp == NULL)
	  panic(0,"Missing required group `mbuni' in config file!");

     mms_load_core_settings(cgrp);

     if (proxyrelays)
	  *proxyrelays = mms_proxy_relays(cfg);
          
     if (mms_cfg_get_int(grp, octstr_imm("max-send-threads"), &m->maxthreads) == -1)
	  m->maxthreads = 10;
     
     m->unified_prefix = _mms_cfg_getx(grp, octstr_imm("unified-prefix"));        
     m->local_prefix = _mms_cfg_getx(grp, octstr_imm("local-prefixes"));        
     m->hostname = _mms_cfg_getx(grp, octstr_imm("hostname"));
	
     if (m->hostname == NULL || octstr_len(m->hostname) == 0)
	  m->hostname = octstr_create("localhost");

     m->name = _mms_cfg_getx(grp, octstr_imm("name"));
     m->host_alias = _mms_cfg_getx(grp, octstr_imm("host-alias"));

     m->sendmail = _mms_cfg_getx(grp, octstr_imm("send-mail-prog"));      
     
     qdir = _mms_cfg_getx(grp, octstr_imm("storage-directory"));     
     if (qdir && octstr_len(qdir) >= QFNAMEMAX)
	  warning(0, "storage-directory name too long. Max length is %d", QFNAMEMAX);

     if (mkdir(octstr_get_cstr(qdir), 
	       S_IRWXU|S_IRWXG) < 0 && 
	 errno != EEXIST)
	  panic(0, "Failed to create queue directory: %s - %s!",
		octstr_get_cstr(qdir), strerror(errno));

     m->global_queuedir = octstr_format("%S/global", qdir);
     m->mm1_queuedir = octstr_format("%S/mm1", qdir);
     m->mm4_queuedir = octstr_format("%S/mm4", qdir);


     if (mms_init_queuedir(m->mm1_queuedir) < 0)
	  panic(0, "Failed to initialise local queue directory: %s - %s!", 
		octstr_get_cstr(m->mm1_queuedir), strerror(errno));
     else if (mms_init_queuedir(m->global_queuedir) < 0)
	  panic(0, "Failed to initialise global queue directory: %s - %s!", 
		octstr_get_cstr(m->global_queuedir), strerror(errno));
     else if (mms_init_queuedir(m->mm4_queuedir) < 0)
	  panic(0, "Failed to initialise global queue directory: %s - %s!", 
		octstr_get_cstr(m->mm4_queuedir), strerror(errno));

     m->mmbox_rootdir = octstr_format("%S/mmbox", qdir);
     
     if (mmbox_root_init(octstr_get_cstr(m->mmbox_rootdir)) != 0)
	  panic(0, "Failed to initialise mmbox root directory, error: %s!", 
		strerror(errno));
     
     m->ua_profile_cache_dir = octstr_format("%S/UserAgent_Profiles", qdir);
     
     if (mkdir(octstr_get_cstr(m->ua_profile_cache_dir), 
	       S_IRWXU|S_IRWXG) < 0 && 
	 errno != EEXIST)
	  panic(0, "Failed to initialise UA Profile directory, error: %s!",
		strerror(errno));

     if (mms_cfg_get_int(grp, octstr_imm("maximum-send-attempts"), &m->maxsendattempts) == -1)
	  m->maxsendattempts = MAXQTRIES;

     if (mms_cfg_get_int(grp, octstr_imm("default-message-expiry"), &m->default_msgexpiry) == -1)
	  m->default_msgexpiry = DEFAULT_EXPIRE;

     s = _mms_cfg_getx(grp, octstr_imm("queue-run-interval"));
     if (!s || (m->queue_interval = atof(octstr_get_cstr(s))) <= 0)
	  m->queue_interval = QUEUERUN_INTERVAL;

     if (s)
	octstr_destroy(s);
     
     if (mms_cfg_get_int(grp, octstr_imm("send-attempt-back-off"), &m->send_back_off) == -1)
	  m->send_back_off = BACKOFF_FACTOR;

     /* Make send sms url. */
     m->sendsms_url = _mms_cfg_getx(grp, octstr_imm("sendsms-url"));     
     
     user = _mms_cfg_getx(grp, octstr_imm("sendsms-username"));     
     pass = _mms_cfg_getx(grp, octstr_imm("sendsms-password"));     
     from = mms_cfg_get(grp, octstr_imm("sendsms-global-sender"));     

     if (user && octstr_len(user) > 1)
	  octstr_url_encode(user);
     if (pass && octstr_len(pass) > 1)
	  octstr_url_encode(pass);
     if (from && octstr_len(from) > 1) 
	  octstr_url_encode(from);

     
     i = octstr_search_char(m->sendsms_url, '?', 0); /* If ? is in there, omit below. */
     octstr_format_append(m->sendsms_url, 
			  (from && octstr_len(from) > 1) ? 
			  "%susername=%S&password=%S&from=%S" : 
			  "%susername=%S&password=%S",	  
			  (i >= 0) ? "" : "?",
			  user,
			  pass,from);	
   
     m->system_user = octstr_format("system-user@%S", 
				    m->hostname);	  
     octstr_destroy(user);
     octstr_destroy(pass);
     if (from) octstr_destroy(from);
     
     mms_cfg_get_int(grp, octstr_imm("mms-port"), &port);

     m->port = (port > 0) ? port : MMS_PORT;

     m->mm7port = -1;
     mms_cfg_get_int(grp, octstr_imm("mm7-port"), &m->mm7port);

     m->allow_ip = _mms_cfg_getx(grp, octstr_imm("allow-ip"));          
     m->deny_ip = _mms_cfg_getx(grp, octstr_imm("deny-ip"));          

     m->email2mmsrelay_hosts = _mms_cfg_getx(grp, 
					   octstr_imm("email2mms-relay-hosts"));
     
     m->prov_notify = _mms_cfg_getx(grp,octstr_imm("prov-server-notify-script"));

     m->prov_getstatus = _mms_cfg_getx(grp,octstr_imm("prov-server-sub-status-script"));     
     m->mms_notify_txt = _mms_cfg_getx(grp, octstr_imm("mms-notify-text"));
     m->mms_notify_unprov_txt = _mms_cfg_getx(grp, octstr_imm("mms-notify-unprovisioned-text"));


     m->mms_email_txt = _mms_cfg_getx(grp, octstr_imm("mms-to-email-txt"));
     m->mms_email_html = _mms_cfg_getx(grp, octstr_imm("mms-to-email-html"));
     m->mms_email_subject = mms_cfg_get(grp, octstr_imm("mms-to-email-default-subject"));

     m->mms_toolarge = _mms_cfg_getx(grp, octstr_imm("mms-message-too-large-txt"));

     m->wap_gw_msisdn_header = mms_cfg_get(grp, octstr_imm("mms-client-msisdn-header"));
     if (!m->wap_gw_msisdn_header) m->wap_gw_msisdn_header = octstr_imm(XMSISDN_HEADER);

     m->wap_gw_ip_header = mms_cfg_get(grp, octstr_imm("mms-client-ip-header"));
     if (!m->wap_gw_ip_header) m->wap_gw_ip_header = octstr_imm(XIP_HEADER);

     mms_cfg_get_bool(grp, octstr_imm("notify-unprovisioned"), &m->notify_unprovisioned);

     m->billing_params = _mms_cfg_getx(grp, 
				  octstr_imm("billing-module-parameters"));
     /* Get and load the billing lib if any. */
     
     if ((m->mms_billfuncs = load_module(grp, "billing-library",  "mms_billfuncs"))) {
	  if (m->mms_billfuncs->mms_billingmodule_init == NULL ||
	      m->mms_billfuncs->mms_billmsg == NULL ||
	      m->mms_billfuncs->mms_billingmodule_fini == NULL ||
	      m->mms_billfuncs->mms_logcdr == NULL) 
	       panic(0, "Missing or NULL functions in billing module!");
     } else 
	  m->mms_billfuncs = &mms_billfuncs; /* The default one. */	  
     
     m->mms_bill_module_data = m->mms_billfuncs->mms_billingmodule_init(octstr_get_cstr(m->billing_params));
     
     m->resolver_params = _mms_cfg_getx(grp, 
				   octstr_imm("resolver-module-parameters"));
     
     /* Get and load the resolver lib if any. */
     if ((m->mms_resolvefuncs = load_module(grp, "resolver-library",  "mms_resolvefuncs"))) {
	  if (m->mms_resolvefuncs->mms_resolvermodule_init == NULL ||
	      m->mms_resolvefuncs->mms_resolve == NULL ||
	      m->mms_resolvefuncs->mms_resolvermodule_fini == NULL)
	       panic(0, "Missing or NULL functions in resolver module!");
     } else 
	  m->mms_resolvefuncs = &mms_resolvefuncs;	/* The default one. */

     m->mms_resolver_module_data = m->mms_resolvefuncs->mms_resolvermodule_init(octstr_get_cstr(m->resolver_params));

     m->detokenizer_params = _mms_cfg_getx(grp, octstr_imm("detokenizer-module-parameters"));
     
     /* Get and load the detokenizer lib if any. */
     if ((m->mms_detokenizefuncs = load_module(grp, "detokenizer-library", "mms_detokenizefuncs"))) {
	  if (m->mms_detokenizefuncs->mms_detokenizer_init == NULL ||
	      m->mms_detokenizefuncs->mms_detokenize == NULL ||
	      m->mms_detokenizefuncs->mms_gettoken == NULL ||
	      m->mms_detokenizefuncs->mms_detokenizer_fini == NULL)
	       panic(0, "Missing or NULL functions in detokenizer module!");
	  if (m->mms_detokenizefuncs->mms_detokenizer_init(octstr_get_cstr(m->detokenizer_params)))
	      panic(0, "Detokenizer module failed to initialize");
     } else 
	  m->mms_detokenizefuncs = NULL;

     if (mms_cfg_get_bool(grp, octstr_imm("allow-ip-type"), &m->allow_ip_type)  < 0)
	  m->allow_ip_type = 1;
     
     mms_cfg_get_bool(grp, octstr_imm("optimize-notification-size"), 
		      &m->optimize_notification_size);

     if (mms_cfg_get_bool(grp, octstr_imm("content-adaptation"), &m->content_adaptation) < 0)
	  m->content_adaptation = 1;
     
     if (qdir)
	  octstr_destroy(qdir);
     
     /* Now load the VASP list. */
     l = mms_cfg_get_multi(cfg, octstr_imm("mms-vasp"));
     m->vasp_list = list_create();
     for (i=0, n=list_len(l); i<n; i++) {
	  mCfgGrp *grp = list_get(l, i);
	  MmsVasp *mv = gw_malloc(sizeof *mv);
	  Octstr *s;
	  int ibool = 0;
	  
	  mv->id = _mms_cfg_getx(grp, octstr_imm("vasp-id"));	  
	  mv->short_code = -1;
	  mms_cfg_get_int(grp, octstr_imm("short-code"), &mv->short_code);

	  mv->vasp_username = _mms_cfg_getx(grp, octstr_imm("vasp-username"));
	  mv->vasp_password = _mms_cfg_getx(grp, octstr_imm("vasp-password"));
	  
	  mv->vasp_url = _mms_cfg_getx(grp, octstr_imm("vasp-url"));
	  
	  s = _mms_cfg_getx(grp, octstr_imm("type"));

	  if (octstr_case_compare(s, octstr_imm("soap")) == 0)
	       mv->type = SOAP_VASP;
	  else if (octstr_case_compare(s, octstr_imm("eaif")) == 0)
	       mv->type = EAIF_VASP;
	  else
	       mv->type = NONE_VASP;
	  octstr_destroy(s);
	  
	  /* Set the handler vasp accounts. */
	  if (mms_cfg_get_bool(grp, octstr_imm("mms-to-email-handler"), &ibool) == 0 &&
	      ibool) {
	       if (m->mms2email)
		    warning(0, "mms-to-email handler VASP specified more than once! Only last config taken.");
	       m->mms2email = mv;
	  }	  
	  if (mms_cfg_get_bool(grp, octstr_imm("mms-to-local-copy-handler"), &ibool) == 0 &&
	      ibool) {
	       if (m->mms2mobile)
		    warning(0, "mms-to-mobile copy handler VASP specified more than once! Only last config taken."); 
	       m->mms2mobile = mv;
	  }
	  list_append(m->vasp_list, mv);
     }
     list_destroy(l, NULL);
     return m;
}

List *mms_proxy_relays(mCfg *cfg)
{
     List *gl = mms_cfg_get_multi(cfg, octstr_imm("mmsproxy"));
     int i, n;
     List *l = list_create();
     
     for (i = 0, n = list_len(gl); i < n; i++) {
	  mCfgGrp *grp = list_get(gl, i);
	  MmsProxyRelay *m = gw_malloc(sizeof *m);
	  
	  m->host = _mms_cfg_getx(grp, octstr_imm("host"));
	  m->name = _mms_cfg_getx(grp, octstr_imm("name"));
	  m->allowed_prefix = _mms_cfg_getx(grp, octstr_imm("allowed-prefix"));
	  m->denied_prefix = _mms_cfg_getx(grp, octstr_imm("denied-prefix"));
	  
	  list_append(l, m);
     }
     
     list_destroy(gl, NULL);

     return l;
}

Octstr *mms_makefetchurl(char *qf, Octstr *token, int loc,
			 Octstr *to,
			 MmscSettings *settings)
{
     Octstr *url = octstr_create("");
     Octstr *host_alias = settings->host_alias;
     Octstr *hstr; 
     Octstr *endtoken, *x; 

     MmsDetokenizerFuncStruct *tfs = settings->mms_detokenizefuncs;
     
     if (host_alias && octstr_len(host_alias) > 0)
	  hstr = octstr_duplicate(host_alias);
     else 
	  hstr = octstr_format("%S:%d",
			       settings->hostname, settings->port);
     
     octstr_format_append(url, "http://%S/%s@%d", 
			  hstr, 
			  qf, loc);
     
     if (tfs && tfs->mms_gettoken) { /* we append the recipient token or we append the message token. */
	  endtoken =  tfs->mms_gettoken(to); 
	  if (!endtoken) 
	       endtoken = octstr_imm("x");
     } else {
	  if (!token) 
	       endtoken = octstr_imm("x");	  
	  else
	       endtoken = octstr_duplicate(token);
     }
     
     x = octstr_duplicate(endtoken); /* might be immutable, so we duplicate it. */
     octstr_url_encode(x);
     octstr_format_append(url, "/%S", x);

     octstr_destroy(endtoken);
     octstr_destroy(x);
     octstr_destroy(hstr);
     return url;
}


Octstr *mms_find_sender_msisdn(Octstr *send_url, 
			       Octstr *ip,
			       List *request_hdrs, 
			       Octstr *msisdn_header, 
			       Octstr *requestip_header,
			       MmsDetokenizerFuncStruct* detokenizerfuncs)
{
     /* Either we have a WAP gateway header as defined, or we look for 
      * last part of url, pass it to detokenizer lib if defined, and back comes our number.
      */
     
     Octstr *phonenum = http_header_value(request_hdrs, 
					  msisdn_header);
     
     if (!phonenum || octstr_len(phonenum) == 0) {
	  List *l  = octstr_split(send_url, octstr_imm("/"));
	  Octstr *xip = http_header_value(request_hdrs, 
					  requestip_header);
	  if (xip == NULL)
	       xip = ip ? octstr_duplicate(ip) : NULL;
	  if (detokenizerfuncs && ((l && list_len(l) > 1) || xip)) 
	       phonenum = detokenizerfuncs->mms_detokenize((l && list_len(l) > 1) ? 
							   list_get(l, list_len(l) - 1) :
							   send_url, 
							   xip);	  
	  if (l)
	       list_destroy(l, (list_item_destructor_t *)octstr_destroy);
	  if (xip)
	       octstr_destroy(xip);
     }
     
     return phonenum;     
}

Octstr *mms_find_sender_ip(List *request_hdrs, Octstr *ip_header, Octstr *ip, int *isv6)
{
     Octstr *xip;
     /* Look in the headers, if none is defined, return actual IP */
     Octstr *client_ip = http_header_value(request_hdrs, ip_header);
     char *s;
     
     xip = client_ip  ? client_ip : octstr_duplicate(ip);
     
     s = octstr_get_cstr(xip);

     /* Crude test for ipv6 */
     *isv6 = (index(s, ':') != NULL);
     return xip;
}

int mms_decodefetchurl(Octstr *fetch_url,  
		       Octstr **qf, Octstr **token, int *loc)
{
     Octstr *xfurl = octstr_duplicate(fetch_url);
     int i, j, n;
     char *s, *p;
     
     for (i = 0, n = 0, s = octstr_get_cstr(xfurl); 
	  i < octstr_len(xfurl); i++)
	  if (s[i] == '/')
	       n++;
     if (n < 2) /* We need at least two slashes. */
	  octstr_append_char(xfurl, '/');
     
     i = 0;
     n = octstr_len(xfurl);
     s = octstr_get_cstr(xfurl);
     
     p = strrchr(s, '/'); /* Find last slash. */
     if (p)
	  i = (p - s) - 1;
     else
	  i = n-1;
     
     if (i < 0)
	  i = 0;
     
     while (i>0 && s[i] != '/')
	  i--; /* Go back, find first slash */
     if (i>=0 && s[i] == '/')
	  i++;
     
     /* Now we have qf, find its end. */
     
     j = i;     
     while (j<n && s[j] != '/')
	  j++; /* Skip to next slash. */
     
     *qf = octstr_copy(fetch_url, i, j-i);
     
     if (j<n)
	  *token = octstr_copy(fetch_url, j + 1, n - (j+1));
     else
	  *token = octstr_create("");
     octstr_destroy(xfurl);

     /* Now get loc out of qf. */
     *loc = MMS_LOC_MQUEUE;
     i = octstr_search_char(*qf, '@', 0);
     if (i >= 0) {
	  long l;
	  int j = octstr_parse_long(&l, *qf, i+1, 10);
	  if (j > 0)
	       *loc = l;
	  octstr_delete(*qf, i, octstr_len(*qf));
     } 
     
     return 0;
}


int mms_ind_send(Octstr *prov_cmd, Octstr *to)
{
     Octstr *tmp;
     Octstr *s;
     int res = 1;
     
     if (prov_cmd == NULL ||
	 octstr_len(prov_cmd) == 0)
	  return 1;
     
     tmp = octstr_duplicate(to);
     escape_shell_chars(tmp);     
     s = octstr_format("%S %S", prov_cmd, tmp); 
     octstr_destroy(tmp);

     if (s) {
	  int x = system(octstr_get_cstr(s));
	  int y = WEXITSTATUS(x);

	  if (x < 0) {
	       error(0, "Checking MMS Ind.Send: Failed to run command %s!", 
		     octstr_get_cstr(s));
	       res = 1;	
	  } else if (y != 0 && y != 1)
	       res =  -1;
	  else 
	       res = y;
	  octstr_destroy(s);
     } else 
	  warning(0, "Checking MMS Ind.Send: Failed call to compose command [%s] ", 
	       octstr_get_cstr(prov_cmd));

	  
     return res;
}

void notify_prov_server(char *cmd, char *from, char *event, char *arg)
{
     Octstr *s;
     Octstr *tmp;
     
     if (cmd == NULL || cmd[0] == '\0')
	  return;
     tmp = octstr_create(from);
     escape_shell_chars(tmp);     
     s = octstr_format("%s '%s' '%s' '%s'", cmd, event, octstr_get_cstr(tmp), arg); 
     octstr_destroy(tmp);

     if (s) {
	  system(octstr_get_cstr(s));	  
	  octstr_destroy(s);
     }
}
